#ifndef	DMRAID_LIB_VERSION

#define	DMRAID_LIB_VERSION		"1"

#define	DMRAID_LIB_MAJOR_VERSION	1
#define	DMRAID_LIB_MINOR_VERSION	0
#define	DMRAID_LIB_SUBMINOR_VERSION	0
#define	DMRAID_LIB_VERSION_SUFFIX	"rc16-4"

#define	DMRAID_LIB_DATE			"(2010.02.17)"

#endif
